/*
 * parse.c
 * David Moore <dcm@csail.mit.edu>
 *
 * Library for parsing raw cricket output
 *
 *
 * Copyright (C) 2004  Massachusetts Institute of Technology
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <sys/time.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "parse.h"

void process_line_v1(parse_info * parse, event_queue * events,
        char * str, long long linetime);
void process_line_v2(parse_info * parse, event_queue * events,
        char * str, long long linetime);

void us_to_tv(struct timeval * tv, long long us)
{
    tv->tv_sec = us / 1000000;
    tv->tv_usec = us % 1000000;
}

long long diff_time(struct timeval * tv1, struct timeval * tv2)
{
    return ((long long)tv1->tv_sec - tv2->tv_sec)*1000000 +
        tv1->tv_usec - tv2->tv_usec;
}

event_queue * init_queue()
{
    event_queue * events;

    events = (event_queue *) malloc(sizeof(event_queue));
    if (!events)
        return NULL;

    events->head = NULL;
    events->tail = NULL;

    return events;
}

void free_queue(event_queue * events)
{
    if (events->head)
        fprintf(stderr, "Warning: free_queue() freeing non-empty queue\n");
    free(events);
}

event * dequeue(event_queue * events)
{
    event * e;

    if (!events->head)
        return NULL;

    e = events->head;
    events->head = e->next;
    if (!events->head) {
        events->tail = NULL;
    }
    return e;
}

void enqueue(event_queue * events, event * e)
{
    event ** enextptr, * e1;

    if (!events->head) {
        events->head = e;
        events->tail = e;
        e->next = NULL;
        return;
    }

    if (diff_time(&e->time, &events->tail->time) >= 0) {
        events->tail->next = e;
        events->tail = e;
        e->next = NULL;
        return;
    }

    enextptr = &events->head;
    e1 = events->head;
    while (diff_time(&e->time, &e1->time) >= 0) {
        enextptr = &e1->next;
        e1 = e1->next;
    }

    *enextptr = e;
    e->next = e1;
}

int queue_empty(event_queue * events)
{
    if (events->head)
        return 0;
    else
        return 1;
}

long long next_event_time(event_queue * events)
{
    if (!events->head)
        return 0;

    return events->head->time.tv_sec * 1000000 + events->head->time.tv_usec;
}

int dispatch_event(parse_info * parse, event * e)
{
    int res;
    switch (e->type) {
        case EVENT_RANGE:
            if (parse->cb_range)
                res = parse->cb_range((event_range *) e, &e->time);
            break;
        case EVENT_NEWNODE:
            if (parse->cb_newnode)
                res = parse->cb_newnode((event_newnode *) e, &e->time);
            break;
        case EVENT_LOCALIZE:
            if (parse->cb_localize)
                res = parse->cb_localize((event_localize *) e, &e->time);
            break;
    }
    return res;
}

void free_event(event * e)
{
    if (e->type == EVENT_LOCALIZE) {
        event_localize * el = (event_localize *) e;
        free(el->pos);
        free(el->quad);
    }
    free(e);
}

int dispatch_timer(parse_info * parse, long long elapsed)
{
    struct timeval tv;

    if (!parse->cb_timer)
        return 0;

    us_to_tv(&tv, elapsed);
    return parse->cb_timer(&tv);
}

void fetch_next_line(parse_info * parse, long long * nexttime)
{
    int c = 0;
    int sec, msec;

    while (fgets(parse->linebuf, sizeof(parse->linebuf), parse->infile)) {
        c = sscanf(parse->linebuf, "%d.%d:", &sec, &msec);
        if (c == 2) {
            parse->hastime = 1;
            break;
        }
        if (parse->hastime == -1 && c == 0 &&
                !strncmp(parse->linebuf, "node", 4)) {
            parse->hastime = 0;
        }
        if (parse->hastime == 0) {
            if (!strncmp(parse->linebuf, "TX:", 3)) {
                *nexttime = *nexttime + 1000000;
            }
            return;
        }
    }
    if (feof(parse->infile)) {
        parse->has_lines = 0;
        return;
    }

    *nexttime = (long long)sec * 1000000 + msec * 1000 - parse->start_line_time;
}

void process_line(parse_info * parse, event_queue * events,
        char * str, long long linetime)
{
    struct timeval tv;
    if (parse->verbose) {
        us_to_tv(&tv, linetime);
        fprintf(parse->verbose, "%04ld.%03ld: %s\n",
                tv.tv_sec, tv.tv_usec/1000, str);
    }
    if (!parse->line) {
        if (!strncmp(str, "node ", 5)) {
            parse->line = process_line_v2;
            if (parse->verbose)
                fprintf(parse->verbose, "Detected data format 2\n");
        }
        else {
            parse->line = process_line_v1;
            if (parse->verbose)
                fprintf(parse->verbose, "Detected data format 1\n");
        }
    }
    parse->line(parse, events, str, linetime);
}

void parse_next_line(parse_info * parse, event_queue * events,
        long long elapsed)
{
    if (parse->use_serial) {
        int len;
        char * s, * a;
        char buf[256];

        len = read(parse->serfd, buf, sizeof(buf)-1);
        buf[len] = '\0';
        s = buf;
        while ((a = strchr(s, '\n'))) {
            *a = '\0';
            strcat(parse->linebuf, s);
            process_line(parse, events, parse->linebuf, elapsed);
            strcpy(parse->linebuf, "");
            s = a+1;
        }
        strcat(parse->linebuf, s);
    }
    else {
        parse->linebuf[strlen(parse->linebuf)-1] = '\0';
        if (parse->hastime == 0)
            process_line(parse, events, parse->linebuf,
                    parse->next_line_time);
        else
            process_line(parse, events, strchr(parse->linebuf,' ')+1,
                    parse->next_line_time);
        fetch_next_line(parse, &parse->next_line_time);
    }
}

int wait_for_action(event_queue * events, long long next_timer,
        parse_info * parse, long long * elapsed)
{
    int next_type = ACTION_TIMER;
    long long next = next_timer;
    struct timeval now;

    if (!queue_empty(events) && next_event_time(events) <= next) {
        next_type = ACTION_EVENT;
        next = next_event_time(events);
    }

    if (parse->use_serial && next > *elapsed + 10000) {
        fd_set rfds;
        int ret;
        struct timeval timeout;

        FD_ZERO(&rfds);
        FD_SET(parse->serfd, &rfds);
        us_to_tv(&timeout, next - *elapsed - 10000);
        ret = select(parse->serfd+1, &rfds, NULL, NULL, &timeout);
        if (ret == 1)
            next_type = ACTION_PARSE;
        gettimeofday(&now, NULL);
        *elapsed = diff_time(&now, &parse->starttime);
    }
    else if (!parse->use_serial) {
        if (parse->has_lines &&
                parse->next_line_time < next + parse->readahead_time) {
            next_type = ACTION_PARSE;
            next = parse->next_line_time - parse->readahead_time;
        }

        if (parse->realtime && next > *elapsed + 10000) {
            usleep(next - *elapsed - 10000);
            gettimeofday(&now, NULL);
            *elapsed = diff_time(&now, &parse->starttime);
        }
        else if (!parse->realtime) {
            *elapsed = next;
        }
    }

    return next_type;
}


parse_info * parser_new()
{
    parse_info * parse;

    parse = (parse_info *) malloc(sizeof(parse_info));
    if (!parse)
        return NULL;

    parse->cb_range = NULL;
    parse->cb_newnode = NULL;
    parse->cb_localize = NULL;
    parse->cb_timer = NULL;
    parse->line = NULL;
    parse->hastime = -1;
    return parse;
}

parse_info * parser_init_serial(int fd, FILE * verbose)
{
    parse_info * parse;

    parse = parser_new();
    if (!parse)
        return NULL;

    parse->use_serial = 1;
    parse->serfd = fd;
    parse->verbose = verbose;
    return parse;
}

parse_info * parser_init_file(FILE * infile, int realtime, int readahead_time,
        FILE * verbose)
{
    parse_info * parse;

    parse = parser_new();
    if (!parse)
        return NULL;

    parse->use_serial = 0;
    parse->infile = infile;
    parse->realtime = realtime;
    parse->readahead_time = readahead_time;
    parse->verbose = verbose;
    return parse;
}

void parser_free(parse_info * parse)
{
    free(parse);
}

int parser_mainloop(parse_info * parse)
{
    long long next_timer = 50000;
    event_queue * events;
    long long elapsed = 0;
    int i, res = 0;

    events = init_queue();

    parse->event_buf = NULL;
    parse->selfnode = -1;
    for (i=0; i<256; i++)
        parse->v1_nodes[i] = 0;

    parse->start_line_time = 0;
    parse->next_line_time = 0;
    parse->has_lines = 1;
    strcpy(parse->linebuf, "");
    fetch_next_line(parse, &parse->start_line_time);

    gettimeofday(&parse->starttime, NULL);
    while (1) {
        event * e;

        switch (wait_for_action(events, next_timer, parse, &elapsed)) {
            case ACTION_EVENT:
                e = dequeue(events);
                res = dispatch_event(parse, e);
                free_event(e);
                break;
            case ACTION_TIMER:
                res = dispatch_timer(parse, elapsed);
                next_timer += 50000;
                break;
            case ACTION_PARSE:
                parse_next_line(parse, events, elapsed);
                break;
        }

        if (res < 0)
            break;
        if (!parse->use_serial && !parse->has_lines && queue_empty(events))
            break;
    }
    free_queue(events);
    return 0;
}


void process_dists_v2(event_queue * events, int node, char * str,
        long long linetime)
{
    int j, c;
    int ret, dist;

    while (1) {
        event * e;
        event_range * er;
        ret = sscanf(str, " %d d=%d ;%n", &j, &dist, &c);
        str += c;
        if (ret != 2)
            break;
        e = (event *) malloc(sizeof(event_range));
        er = (event_range *) e;
        e->type = EVENT_RANGE;
        us_to_tv(&e->time, linetime);
        er->summary = 1;
        er->dist = dist;
        er->node_send = j;
        er->node_recv = node;
        enqueue(events, e);
    }
}

void process_line_v2(parse_info * parse, event_queue * events,
        char * str, long long linetime)
{
    int i, n;
    int ret;
    float x, y;
    
    if (!strcmp(str, "Localizing...")) {
        event_localize * el;
        if (parse->event_buf) {
            enqueue(events, parse->event_buf);
            parse->event_buf = NULL;
        }
        parse->event_buf = (event *) malloc(sizeof(event_localize));
        parse->event_buf->type = EVENT_LOCALIZE;
        us_to_tv(&parse->event_buf->time, linetime);
        el = (event_localize *) parse->event_buf;
        el->method = METHOD_ROBUST_QUADS;
        el->num_opt = 0;
        el->opt = NULL;
        el->num_pos = 0;
        el->pos = NULL;
        el->num_quad = 0;
        el->quad = NULL;
    }
    else if (!strncmp(str, "Tracking", 8)) {
        event_localize * el;
        int t1, t2;
        if (parse->event_buf) {
            enqueue(events, parse->event_buf);
            parse->event_buf = NULL;
        }
        if (sscanf(str, "Tracking %d time=%u,clock=%u", &i, &t1, &t2) != 3)
            return;
        t1 = ((t2 - t1) & 0xffff) * 1000 * (1000.0 / 1024.0);
        if (t1 > linetime)
            return;
        parse->event_buf = (event *) malloc(sizeof(event_localize));
        parse->event_buf->type = EVENT_LOCALIZE;
        us_to_tv(&parse->event_buf->time, linetime - t1);
        el = (event_localize *) parse->event_buf;
        el->method = METHOD_MOBILE_LSQ;
        el->num_opt = 0;
        el->opt = NULL;
        el->num_pos = 0;
        el->pos = NULL;
        el->num_quad = 0;
        el->quad = NULL;
    }
    else if (!strncmp(str, "Rigid Quad:", 11) || !strncmp(str, "RQ:", 3)) {
        int ids[3];
        event_localize * el = (event_localize *) parse->event_buf;
        if (!parse->event_buf || parse->event_buf->type != EVENT_LOCALIZE) {
            return;
        }
        if (sscanf(str, "Rigid Quad: %d %d %d", ids, ids+1, ids+2) != 3)
            sscanf(str, "RQ: %d %d %d", ids, ids+1, ids+2);
        el->quad = (quad *) realloc(el->quad, (el->num_quad+1)*sizeof(quad));
        for (i=0; i<3; i++)
            el->quad[el->num_quad].v[i] = ids[i];
        el->num_quad++;
    }
    else if (!strncmp(str, "Localized ", 10)) {
        parse->state = STATE_POS;
    }
    else if (!strncmp(str, "Optimized,", 10)) {
        parse->state = STATE_OPT;
    }
    else if (sscanf(str, "  %d: x=%f y=%f%n", &i, &x, &y, &n) == 3) {
        event_localize * el = (event_localize *) parse->event_buf;
        if (!parse->event_buf || parse->event_buf->type != EVENT_LOCALIZE) {
            return;
        }
        if (parse->state == STATE_POS) {
            el->pos = (pos *) realloc(el->pos, (el->num_pos+1)*sizeof(pos));
           
            el->pos[el->num_pos].node = i;
            el->pos[el->num_pos].x = x;
            el->pos[el->num_pos].y = y;
            el->pos[el->num_pos].flags = 0;
            if (strchr(str+n, '*'))
                el->pos[el->num_pos].flags |= POS_FLAG_WEAK;
            el->num_pos++;
        }
        else if (parse->state == STATE_OPT) {
            el->opt = (pos *) realloc(el->opt, (el->num_opt+1)*sizeof(pos));
           
            el->opt[el->num_opt].node = i;
            el->opt[el->num_opt].x = x;
            el->opt[el->num_opt].y = y;
            el->opt[el->num_opt].flags = 0;
            if (strchr(str+n, '*'))
                el->opt[el->num_opt].flags |= POS_FLAG_WEAK;
            el->num_opt++;
        }
    }
    else if (!strcmp(str, "done")) {
    }
    else if (!strcmp(str, "opt done")) {
    }
    else if (!strncmp(str, "beacon:", 7)) {
        int dist, time;
        event * e;
        event_range * er;
        ret = sscanf(str, "beacon:%d,dist=%d,time=%d", &i, &dist, &time);
        if (ret != 3) {
            ret = sscanf(str, "beacon:%d,dist=%d,duration=%d",
                    &i, &dist, &time);
            if (ret != 3)
                return;
        }
        if (i == 0xff)
            return;
        e = (event *) malloc(sizeof(event_range));
        er = (event_range *) e;
        e->type = EVENT_RANGE;
        us_to_tv(&e->time, linetime);
        er->summary = 0;
        er->dist = time;
        er->node_send = i;
        er->node_recv = parse->selfnode;
        enqueue(events, e);
    }
    else if (!strncmp(str, "node ", 5)) {
        int id[4];
        event * e;
        event_newnode * en;
        e = (event *) malloc(sizeof(event_newnode));
        en = (event_newnode *) e;
        e->type = EVENT_NEWNODE;
        us_to_tv(&e->time, linetime);
        if (sscanf(str, "node %d (self) ID %x:%x:%x:%x",
                    &i, id, id+1, id+2, id+3) == 5) {
            en->self = 1;
            parse->selfnode = i;
        }
        else if (sscanf(str, "node %d ID %x:%x:%x:%x",
                    &i, id, id+1, id+2, id+3) == 5) {
            en->self = 0;
        }
        else {
            free_event(e);
            return;
        }
        en->node = i;
        en->id[0] = id[0];
        en->id[1] = id[1];
        en->id[2] = id[2];
        en->id[3] = id[3];
        enqueue(events, e);
    }
    else if (!strncmp(str, "  from ", 7)) {
        char buf[16];
        ret = sscanf(str, " from %d%7c:", &i, buf);
        if (ret == 2) {
            parse->last_node = i;
            process_dists_v2(events, i, str+17, linetime);
        }
    }
    else if (!strncmp(str, "          ", 10)) {
        process_dists_v2(events, parse->last_node, str, linetime);
    }
}

int check_node_v1(parse_info * parse, event_queue * events, int node,
        int isself, long long linetime)
{
    if (parse->selfnode == -1 && !isself)
        return -1;

    if (!parse->v1_nodes[node]) {
        event * e;
        event_newnode * en;
        e = (event *) malloc(sizeof(event_newnode));
        en = (event_newnode *) e;
        e->type = EVENT_NEWNODE;
        us_to_tv(&e->time, linetime);
        en->node = node;
        if (isself) {
            en->self = 1;
            parse->selfnode = node;
        }
        else {
            en->self = 0;
        }
        en->id[0] = 0;
        en->id[1] = 0;
        en->id[2] = 0;
        en->id[3] = node;
        enqueue(events, e);
        parse->v1_nodes[node] = 1;
    }
    return 0;
}

void process_dists_v1(parse_info * parse, event_queue * events, int node,
        char * str, long long linetime)
{
    int j, c;
    int id, dist;

    if (check_node_v1(parse, events, node, 0, linetime))
        return;

    while (1) {
        event * e;
        event_range * er;
        j = sscanf(str, " 0x%x d=%d ;%n", &id, &dist, &c);
        str += c;
        if (j != 2)
            break;
        if (check_node_v1(parse, events, id, 0, linetime))
            continue;
        e = (event *) malloc(sizeof(event_range));
        er = (event_range *) e;
        e->type = EVENT_RANGE;
        us_to_tv(&e->time, linetime);
        er->summary = 1;
        er->dist = dist;
        er->node_send = id;
        er->node_recv = node;
        enqueue(events, e);
    }
}


void process_line_v1(parse_info * parse, event_queue * events,
        char * str, long long linetime)
{
    int i;
    int ret;
    float x, y;
    
    if (!strcmp(str, "Localizing...")) {
        event_localize * el;
        if (parse->event_buf) {
            fprintf(stderr, "Warning: localization started with unfinished event\n");
            return;
        }
        parse->event_buf = (event *) malloc(sizeof(event_localize));
        parse->event_buf->type = EVENT_LOCALIZE;
        el = (event_localize *) parse->event_buf;
        el->num_opt = 0;
        el->opt = NULL;
        el->num_pos = 0;
        el->pos = NULL;
        el->num_quad = 0;
        el->quad = NULL;
    }
    else if (!strncmp(str, "Rigid Quad:", 11)) {
        int ids[3];
        event_localize * el = (event_localize *) parse->event_buf;
        if (!parse->event_buf || parse->event_buf->type != EVENT_LOCALIZE) {
            return;
        }
        sscanf(str, "Rigid Quad: %x %x %x", ids, ids+1, ids+2);
        if (check_node_v1(parse, events, ids[0], 0, linetime) ||
                check_node_v1(parse, events, ids[1], 0, linetime) ||
                check_node_v1(parse, events, ids[2], 0, linetime))
            return;
        el->quad = (quad *) realloc(el->quad, (el->num_quad+1)*sizeof(quad));
        for (i=0; i<3; i++)
            el->quad[el->num_quad].v[i] = ids[i];
        el->num_quad++;
    }
    else if (!strncmp(str, "Localized ", 10)) {
        parse->state = STATE_POS;
    }
    else if (!strncmp(str, "Optimized,", 10)) {
        parse->state = STATE_OPT;
    }
    else if (!strncmp(str, "  0x", 4)) {
        event_localize * el = (event_localize *) parse->event_buf;
        if (!parse->event_buf || parse->event_buf->type != EVENT_LOCALIZE) {
            return;
        }
        if (sscanf(str, "  0x%x: x=%f y=%f", &i, &x, &y) != 3)
            return;

        if (check_node_v1(parse, events, i, 0, linetime))
            return;

        if (parse->state == STATE_POS) {
            el->pos = (pos *) realloc(el->pos, (el->num_pos+1)*sizeof(pos));
           
            el->pos[el->num_pos].node = i;
            el->pos[el->num_pos].x = x;
            el->pos[el->num_pos].y = y;
            el->num_pos++;
        }
        else if (parse->state == STATE_OPT) {
            el->opt = (pos *) realloc(el->opt, (el->num_opt+1)*sizeof(pos));
           
            el->opt[el->num_opt].node = i;
            el->opt[el->num_opt].x = x;
            el->opt[el->num_opt].y = y;
            el->num_opt++;
        }
    }
    else if (!strcmp(str, "done")) {
        if (!parse->event_buf || parse->event_buf->type != EVENT_LOCALIZE) {
            return;
        }
        us_to_tv(&parse->event_buf->time, linetime);
        enqueue(events, parse->event_buf);
        parse->event_buf = NULL;
    }
    else if (!strcmp(str, "opt done")) {
    }
    else if (!strncmp(str, "id=", 3)) {
        int dist, time;
        int myid;
        event * e;
        event_range * er;
        ret = sscanf(str, "id=%x,beacon=%x,dist=%d,time=%d",
                &myid, &i, &dist, &time);
        if (ret != 4)
            return;
        if (check_node_v1(parse, events, myid, 1, linetime))
            return;
        if (check_node_v1(parse, events, i, 0, linetime))
            return;
        e = (event *) malloc(sizeof(event_range));
        er = (event_range *) e;
        e->type = EVENT_RANGE;
        us_to_tv(&e->time, linetime);
        er->summary = 0;
        er->dist = time;
        er->node_send = i;
        er->node_recv = parse->selfnode;
        enqueue(events, e);
    }
    else if (!strncmp(str, "  from 0x", 9)) {
        char buf[16];
        ret = sscanf(str, " from 0x%x%7c:", &i, buf);
        if (ret == 2) {
            if (!strcmp(buf, " (self)"))
                check_node_v1(parse, events, i, 1, linetime);
            parse->last_node = i;
            process_dists_v1(parse, events, i, str+19, linetime);
        }
    }
    else if (!strncmp(str, "          0x", 12)) {
        process_dists_v1(parse, events, parse->last_node, str, linetime);
    }
}
