/*
 * sim_all_clusters.c
 * David Moore <dcm@csail.mit.edu>
 *
 *
 *
 * Copyright (C) 2005  Massachusetts Institute of Technology
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>

#include "parse.h"

#define MAX_NODES   50

int ranges_total = 0;
int num_ranges = 0;
int full_count = 0;
int node_ids[MAX_NODES];
FILE * outfile;

int range(event_range * e, struct timeval * tv)
{
    if (e->summary)
        ranges_total++;
    return 0;
}

int newnode(event_newnode * e, struct timeval * tv)
{
    if (e->self)
        return 0;

    node_ids[e->node] = (e->id[2] << 8) | e->id[3];
    return 0;
}

int localize(event_localize * e, struct timeval * tv)
{
    if (e->method == METHOD_ROBUST_QUADS) {
        printf("%d ", e->num_pos);
        fflush(stdout);
        if (ranges_total == num_ranges)
            full_count++;
        ranges_total = 0;
        if (full_count == 3) {
            int i;
            full_count = 0;
            for (i=0; i<e->num_pos; i++) {
                fprintf(outfile, "%d %f %f %d\n", node_ids[e->pos[i].node], e->pos[i].x,
                        e->pos[i].y, (e->pos[i].flags & POS_FLAG_WEAK)?1:0);
            }
            return -1;
        }
    }
    return 0;
}

FILE * sim_open(char * filename, int num_nodes, int * pid, char * cmd)
{
    int p[2], proc;

    if (pipe(p) < 0)
        return NULL;
    
    proc = fork();
    if (proc > 0) { /* parent */
        *pid = proc;
        close(p[1]);
        return fdopen(p[0], "r");
    }

    if (proc == 0) {
        char infilearg[256];
        char num_str[16];
        close(1);
        dup(p[1]);
        close(p[1]);
        close(p[0]);
        snprintf(infilearg, sizeof(infilearg), "-d=%s", filename);
        snprintf(num_str, sizeof(num_str), "%d", num_nodes);
        execl(cmd, "main.exe", "-b=0", infilearg, num_str, NULL);
        perror("execl");
        exit(1);
    }

    return NULL;
}

void sim_close(FILE * file, int pid)
{
    kill(pid, SIGTERM);
    wait4(pid, NULL, 0, NULL);
    fclose(file);
}

int main(int argc, char ** argv)
{
    char c;
    FILE * input = NULL;
    FILE * idxfile;
    parse_info * parse;
    int n=0;
    char * dirname = NULL;
    char * simcmd = NULL;
    char infile[256];
    char str[256];
    int pid;

    while ((c = getopt(argc, argv, "i:c:")) >= 0) {
        switch (c) {
            case 'i':
                dirname = strdup(optarg);
                break;
            case 'c':
                simcmd = strdup(optarg);
                break;
            default:
                exit(1);
        }
    }

    if (!dirname || !simcmd)
        exit(1);

    setenv("DBG", "none", 1);

    snprintf(infile, sizeof(infile), "%s/index", dirname);
    idxfile = fopen(infile, "r");
    if (!idxfile)
        exit(1);
    
    while (fgets(str, sizeof(str), idxfile)) {
        char filename[32];
        char * a;
        FILE * dbgfile;

        if (sscanf(str, "%32s %d %d", filename, &n, &num_ranges) != 3)
            break;
        snprintf(infile, sizeof(infile), "%s/%s", dirname, filename);

        printf("Processing \"%s\" (%d nodes)\n", filename, n);
        
        input = sim_open(infile, n, &pid, simcmd);
        if (!input) {
            perror("sim_open");
            continue;
        }

        a = strstr(filename, ".in");
        if (a)
            *a = '\0';
        snprintf(infile, sizeof(infile), "%s/%s.out", dirname, filename);
        outfile = fopen(infile, "w");
        if (!outfile)
            continue;
        snprintf(infile, sizeof(infile), "%s/%s.dbg", dirname, filename);
        dbgfile = fopen(infile, "w");
        if (!dbgfile)
            continue;

        parse = parser_init_file(input, 0, 0, dbgfile);
        if (!parse)
            continue;
        parse->cb_range = range;
        parse->cb_newnode = newnode;
        parse->cb_localize = localize;

        printf("  Localized: ");
        parser_mainloop(parse);
        printf("\n");
        parser_free(parse);
        sim_close(input, pid);

        fclose(dbgfile);
        fclose(outfile);
    }

    fclose(idxfile);

    return 0;
}
